package ru.dbotthepony.mc.otm.screen.panels;

import com.mojang.blaze3d.systems.RenderSystem;
import com.mojang.blaze3d.vertex.PoseStack;
import net.minecraft.client.gui.components.events.GuiEventListener;
import ru.dbotthepony.mc.otm.screen.MatteryScreen;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;

/**
 * Superclass of all panels
 * This panel represent part of GMod's Panels API
 * which allows much simpler GUI composition
 */
public class EditablePanel implements GuiEventListener {
	public record ScreenPos(float x, float y) {

	}

	private EditablePanel parent;
	public final MatteryScreen<?> screen;

	protected boolean visible_internal = true;
	private boolean visible = true;

	protected boolean focused = false;
	private boolean focused_internal = false;

	// updated from render
	// tells whenever is mouse within this panel's boundaries
	// doesn't tell whenever it is hovered directly above and not on children
	public boolean is_hovered = false;

	// relative to parent
	// updated inside performLayout of parent if docked
	private float x;
	private float y;

	// updated inside performLayout depending on dock
	private float width;
	private float height;

	// updated on each rendered frame
	protected float parent_x;
	protected float parent_y;

	protected float accumulated_depth = 0;

	public float getMost3DHeight() {
		var depth = accumulated_depth;

		for (var child : children) {
			depth = Math.max(child.getMost3DHeight(), depth);
		}

		return depth;
	}

	protected void accumulate3DHeight(float value) {
		accumulated_depth += value;
	}

	public void set3DDepth(float value) {
		accumulated_depth = value;
	}

	private boolean needs_invalidation = false;

	private final ArrayList<EditablePanel> children = new ArrayList<>();

	private Dock docking = Dock.NONE;

	public EditablePanel(@Nonnull MatteryScreen<?> screen, @Nullable EditablePanel parent, float x, float y, float width, float height) {
		this.parent = parent;
		this.screen = screen;
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;

		if (parent != null)
			parent.onParent(this);

		invalidateLayout();
	}

	public float getRenderX() {
		return parent_x;
	}

	public float getRenderY() {
		return parent_y;
	}

	protected void innerRender(PoseStack stack, float mouse_x, float mouse_y, float flag) {

	}

	protected void innerRenderTooltips(PoseStack stack, float mouse_x, float mouse_y, float flag) {

	}

	/**
	 * Called on first pass
	 * background, items, buttons should be drawn from innerRender
	 *
	 * @param stack
	 * @param mouse_x
	 * @param mouse_y
	 * @param flag
	 *
	 * @return float representing the most high depth
	 */
	public float render(PoseStack stack, float mouse_x, float mouse_y, float flag) {
		if (!getVisible()) {
			return 0;
		}

		if (needs_invalidation) {
			needs_invalidation = false;
			performLayout();
		}

		if (parent == null) {
			parent_x = x;
			parent_y = y;

		} else {
			is_hovered = parent.is_hovered && mouse_x >= parent_x && mouse_x <= parent_x + width && mouse_y >= parent_y && mouse_y <= parent_y + height;
		}

		stack.pushPose();
		stack.translate(parent_x, parent_y, accumulated_depth);
		innerRender(stack, mouse_x, mouse_y, flag);
		stack.popPose();

		var most_depth = accumulated_depth;

		for (var child : children) {
			if (child.getVisible()) {
				child.accumulated_depth = accumulated_depth + 1;
				child.parent_x = parent_x + child.x;
				child.parent_y = parent_y + child.y;

				RenderSystem.setShaderColor(1F, 1F, 1F, 1F);
				most_depth = Math.max(most_depth, child.render(stack, mouse_x, mouse_y, flag));
			}
		}

		return most_depth;
	}

	public boolean tickHover(float mouse_x, float mouse_y) {
		return is_hovered = mouse_x >= x && mouse_y >= y && mouse_x <= x + width && mouse_y <= y + height;
	}

	public void unsetHover() {
		is_hovered = false;
	}

	/**
	 * Called on second pass
	 * tooltips and otherwise "on top of everything" should be drawn from innerRenderPost
	 *
	 * @param stack
	 * @param mouse_x
	 * @param mouse_y
	 * @param flag
	 */
	public void renderTooltips(PoseStack stack, float mouse_x, float mouse_y, float flag) {
		if (!getVisible()) {
			return;
		}

		innerRenderTooltips(stack, mouse_x, mouse_y, flag);

		for (var child : children) {
			if (child.getVisible()) {
				child.innerRenderTooltips(stack, mouse_x, mouse_y, flag);
			}
		}
	}

	public void setDocking(Dock docking) {
		if (docking != this.docking) {
			this.docking = docking;

			if (parent != null)
				parent.invalidateLayout();
		}
	}

	public static record DockProperty(float left, float top, float right, float bottom) {
		public boolean equals(Object dock_margin) {
			if (dock_margin instanceof DockProperty prop) {
				return this == prop || left == prop.left && top == prop.top && right == prop.right && bottom == prop.bottom;
			}

			return false;
		}
	}

	public DockProperty getDock_margin() {
		return dock_margin;
	}

	public DockProperty getDock_padding() {
		return dock_padding;
	}

	private DockProperty dock_margin = new DockProperty(0, 0, 0, 0);
	private DockProperty dock_padding = new DockProperty(0, 0, 0, 0);

	public boolean setDockMargin(float left, float top, float right, float bottom) {
		var dock = new DockProperty(left, top, right, bottom);

		if (!dock.equals(dock_margin)) {
			dock_margin = dock;

			if (docking != Dock.NONE && parent != null)
				parent.invalidateLayout();

			return true;
		}

		return false;
	}

	public boolean setDockPadding(float left, float top, float right, float bottom) {
		var dock = new DockProperty(left, top, right, bottom);

		if (!dock.equals(dock_padding)) {
			dock_padding = dock;
			invalidateLayout();
			return true;
		}

		return false;
	}

	public ScreenPos localToScreen(float x, float y) {
		var get_parent = getParent();

		while (get_parent != null) {
			x += get_parent.getX();
			y += get_parent.getY();
			get_parent = get_parent.getParent();
		}

		return new ScreenPos(this.x + x, this.y + y);
	}

	@Nullable
	public EditablePanel getRootParent() {
		if (getParent() == null)
			return null;

		var get_parent = getParent();

		while (true) {
			var get_parent2 = get_parent.getParent();

			if (get_parent2 != null) {
				get_parent = get_parent2;
			} else {
				return get_parent;
			}
		}
	}

	public EditablePanel getRootPanel() {
		var get_parent = getRootParent();
		return get_parent != null ? get_parent : this;
	}

	public ScreenPos localToScreen() {
		return localToScreen(0, 0);
	}

	public ScreenPos screenToLocal(float x, float y) {
		var pos = localToScreen();
		return new ScreenPos(x - pos.x, y - pos.y);
	}

	public ScreenPos screenToLocal(double x, double y) {
		var pos = localToScreen();
		return new ScreenPos((float) (x - pos.x), (float) (y - pos.y));
	}

	public void invalidateLayout() {
		needs_invalidation = true;
	}

	public EditablePanel getParent() {
		return parent;
	}

	public boolean getVisible() {
		return visible_internal && visible;
	}

	public boolean getVisibleThis() {
		return visible;
	}

	protected void visibilityChanges(boolean new_visible, boolean old_visible) {

	}

	private void updateVisible() {
		for (var child : children) {
			var old_visible = child.getVisible();
			child.visible_internal = getVisible();
			var new_visible = child.getVisible();

			if (old_visible != new_visible) {
				child.visibilityChanges(new_visible, old_visible);
			}

			child.updateVisible();
		}
	}

	public void setVisible(boolean visible) {
		if (this.visible != visible) {
			var old_visible = getVisible();
			this.visible = visible;
			var new_visible = getVisible();

			if (old_visible != new_visible) {
				visibilityChanges(new_visible, old_visible);
			}

			updateVisible();
		}
	}

	private boolean hasHierarchicalFocus() {
		return getHierarchicalFocus() != null;
	}

	@Nullable
	private EditablePanel getHierarchicalFocus() {
		if (focused)
			return this;

		for (var child : children) {
			if (child.focused) {
				return child.getHierarchicalFocus();
			}
		}

		return null;
	}

	protected void onFocusChanged(boolean new_focused, boolean old_focused) {

	}

	private void updateFocused() {
		var old_value = focused_internal;
		focused_internal = hasHierarchicalFocus();

		if (focused_internal != old_value) {
			onFocusChanged(focused_internal, old_value);
		}

		for (var child : children) {
			child.updateFocused();
		}
	}

	public boolean getFocused() {
		return focused || focused_internal;
	}

	public boolean getFocusedThis() {
		return focused;
	}

	public void setFocused(boolean focused) {
		if (this.focused != focused) {
			this.focused = focused;
			getRootPanel().updateFocused();
		}
	}

	/**
	 * Focuses this panel
	 */
	public void requestFocus() {
		if (focused)
			return;

		if (focused_internal) {
			var child = getHierarchicalFocus();

			while (child != null) {
				child.setFocused(false);
				child = getHierarchicalFocus();
			}

			setFocused(true);
		} else {
			setFocused(true);
		}
	}

	public void killFocus() {
		if (this.focused_internal) {
			var child = getHierarchicalFocus();

			while (child != null) {
				child.setFocused(false);
				child = getHierarchicalFocus();
			}

			getRootPanel().updateFocused();
		}
	}

	private void onParent(EditablePanel children) {
		this.children.add(children);
		updateVisible();
		invalidateLayout();
	}

	private void onUnparent(EditablePanel children) {
		this.children.remove(children);
		invalidateLayout();
	}

	public void setParent(@Nullable EditablePanel parent) {
		if (this.parent != parent) {
			invalidateLayout();

			if (this.parent != null)
				this.parent.onParent(this);

			this.parent = parent;

			if (parent != null)
				parent.onParent(this);
		}
	}

	protected void xPosUpdated(float new_value, float old_value) {}
	protected void yPosUpdated(float new_value, float old_value) {}

	protected void widthUpdated(float new_value, float old_value) {}
	protected void heightUpdated(float new_value, float old_value) {}

	public void setPos(float x, float y) {
		setX(x);
		setY(y);
	}

	public void setPos(double x, double y) {
		setX((float) x);
		setY((float) y);
	}

	public void setSize(float w, float h) {
		setWidth(w);
		setHeight(h);
	}

	public void setSize(double w, double h) {
		setWidth((float) w);
		setHeight((float) h);
	}

	public void setDimensions(float x, float y, float w, float h) {
		setPos(x, y);
		setSize(w, h);
	}

	public void setDimensions(double x, double y, double w, double h) {
		setPos(x, y);
		setSize(w, h);
	}

	public float getX() {
		return x;
	}

	public void setX(float x) {
		if (x != this.x) {
			invalidateLayout();
			var old = this.x;
			this.x = x;
			xPosUpdated(x, old);
		}
	}

	public float getY() {
		return y;
	}

	public void setY(float y) {
		if (y != this.y) {
			invalidateLayout();
			var old = this.y;
			this.y = y;
			yPosUpdated(y, old);
		}
	}

	public float getWidth() {
		return width;
	}

	public void setWidth(float width) {
		if (width != this.width) {
			invalidateLayout();
			var old = this.width;
			this.width = width;
			widthUpdated(width, old);
		}
	}

	public float getHeight() {
		return height;
	}

	public void setHeight(float height) {
		if (height != this.height) {
			invalidateLayout();
			var old = this.height;
			this.height = height;
			heightUpdated(height, old);
		}
	}

	protected void performLayout() {
		var dock_left = dock_padding.left;
		var dock_right = dock_padding.right;
		var dock_top = dock_padding.top;
		var dock_bottom = dock_padding.bottom;

		// determine sizes
		for (var child : children) {
			switch (child.docking) {
				case TOP -> dock_top += child.height + child.dock_margin.top;
				case BOTTOM -> dock_bottom += child.height + child.dock_margin.bottom;
				case LEFT -> dock_left += child.width + child.dock_margin.left;
				case RIGHT -> dock_right += child.width + child.dock_margin.right;
			}
		}

		var dock_left2 = dock_padding.left;
		var dock_right2 = dock_padding.right;
		var dock_top2 = dock_padding.top;
		var dock_bottom2 = dock_padding.bottom;

		// apply values
		for (var child : children) {
			switch (child.docking) {
				case TOP -> {
					child.setPos(dock_left + child.dock_margin.left, dock_top2 + child.dock_margin.top);
					dock_top2 += child.height + child.dock_margin.top + child.dock_margin.bottom;
					child.setWidth(Math.max(1, width - dock_left - dock_right - child.dock_margin.left - child.dock_margin.right));
				}

				case BOTTOM -> {
					child.setPos(dock_left + child.dock_margin.left, height - dock_bottom2 - child.height - child.dock_margin.bottom);
					dock_bottom2 += child.height + child.dock_margin.top + child.dock_margin.bottom;
					child.setWidth(Math.max(1, width - dock_left - dock_right - child.dock_margin.left - child.dock_margin.right));
				}

				case LEFT -> {
					child.setPos(dock_left2 + child.dock_margin.left, dock_top + child.dock_margin.top);
					dock_left2 += child.width + child.dock_margin.left + child.dock_margin.right;
					child.setHeight(Math.max(1, height - dock_top - dock_bottom - child.dock_margin.top - child.dock_margin.bottom));
				}

				case RIGHT -> {
					child.setPos(width - dock_right2 - child.width - child.dock_margin.right, dock_top + child.dock_margin.top);
					dock_right2 += child.width + child.dock_margin.left + child.dock_margin.right;
					child.setHeight(Math.max(1, height - dock_top - dock_bottom - child.dock_margin.top - child.dock_margin.bottom));
				}
			}
		}
	}

	private boolean accept_mouse_input = true;
	private boolean accept_keyboard_input = true;
	private boolean ignore_mouse_event_boundaries = false;

	public boolean getMouseInputEnabled() {
		return accept_mouse_input;
	}

	public void setMouseInputEnabled(boolean value) {
		accept_mouse_input = value;
	}

	public boolean getKeyboardInputEnabled() {
		return accept_keyboard_input;
	}

	public void setKeyboardInputEnabled(boolean value) {
		accept_keyboard_input = value;
	}

	public boolean getIgnoreMouseEventBoundaries() {
		if (ignore_mouse_event_boundaries) {
			return true;
		}

		for (var child : children) {
			if (child.getIgnoreMouseEventBoundaries()) {
				return true;
			}
		}

		return false;
	}

	// for use when this panel want to attain global focus
	// such as when it is being dragged
	public void setIgnoreMouseEventBoundaries(boolean value) {
		ignore_mouse_event_boundaries = value;
	}

	protected boolean auto_kill_focus = true;

	@Override
	public void mouseMoved(double p_94758_, double p_94759_) {

	}

	protected boolean mouseClickedInner(double mouse_x, double mouse_y, int flag) {
		return false;
	}

	@Override
	public boolean mouseClicked(double mouse_x, double mouse_y, int flag) {
		if (!getVisible() || !getMouseInputEnabled())
			return false;

		if (ignore_mouse_event_boundaries)
			return mouseClickedInner(mouse_x, mouse_y, flag);

		for (var child : children) {
			var pos = child.localToScreen();

			if (child.getIgnoreMouseEventBoundaries() || pos.x <= mouse_x && pos.x + child.width >= mouse_x && pos.y <= mouse_y && pos.y + child.height >= mouse_y) {
				return child.mouseClicked(mouse_x, mouse_y, flag);
			} else if (child.auto_kill_focus) {
				child.killFocus();
			}
		}

		return mouseClickedInner(mouse_x, mouse_y, flag);
	}

	public boolean mouseClickedChecked(double mouse_x, double mouse_y, int flag) {
		if (!getVisible())
			return false;

		var pos = localToScreen();

		if (getIgnoreMouseEventBoundaries() || pos.x <= mouse_x && pos.x + width >= mouse_x && pos.y <= mouse_y && pos.y + height >= mouse_y) {
			if (getMouseInputEnabled() && parent == null)
				screen.popup(this);

			mouseClicked(mouse_x, mouse_y, flag);
			return true;
		}

		return false;
	}

	protected boolean mouseReleasedInner(double mouse_x, double mouse_y, int flag) {
		return false;
	}

	@Override
	public boolean mouseReleased(double mouse_x, double mouse_y, int flag) {
		if (!getVisible() || !getMouseInputEnabled())
			return false;

		if (ignore_mouse_event_boundaries)
			return mouseReleasedInner(mouse_x, mouse_y, flag);

		for (var child : children) {
			var pos = child.localToScreen();

			if (child.getIgnoreMouseEventBoundaries() || pos.x <= mouse_x && pos.x + child.width >= mouse_x && pos.y <= mouse_y && pos.y + child.height >= mouse_y) {
				return child.mouseReleased(mouse_x, mouse_y, flag);
			}
		}

		return mouseReleasedInner(mouse_x, mouse_y, flag);
	}

	public boolean mouseReleasedChecked(double mouse_x, double mouse_y, int flag) {
		if (!getVisible())
			return false;

		var pos = localToScreen();

		if (getIgnoreMouseEventBoundaries() || pos.x <= mouse_x && pos.x + width >= mouse_x && pos.y <= mouse_y && pos.y + height >= mouse_y) {
			mouseReleased(mouse_x, mouse_y, flag);
			return true;
		}

		return false;
	}

	protected boolean mouseDraggedInner(double mouse_x, double mouse_y, int flag, double drag_x, double drag_y) {
		return false;
	}

	@Override
	public boolean mouseDragged(double mouse_x, double mouse_y, int flag, double drag_x, double drag_y) {
		if (!getVisible() || !getMouseInputEnabled())
			return false;

		if (ignore_mouse_event_boundaries)
			return mouseDraggedInner(mouse_x, mouse_y, flag, drag_x, drag_y);

		for (var child : children) {
			var pos = child.localToScreen();

			if (child.getIgnoreMouseEventBoundaries() || pos.x <= mouse_x && pos.x + child.width >= mouse_x && pos.y <= mouse_y && pos.y + child.height >= mouse_y) {
				return child.mouseDragged(mouse_x, mouse_y, flag, drag_x, drag_y);
			}
		}

		return mouseDraggedInner(mouse_x, mouse_y, flag, drag_x, drag_y);
	}

	public boolean mouseDraggedChecked(double mouse_x, double mouse_y, int flag, double drag_x, double drag_y) {
		if (!getVisible())
			return false;

		var pos = localToScreen();

		if (getIgnoreMouseEventBoundaries() || pos.x <= mouse_x && pos.x + width >= mouse_x && pos.y <= mouse_y && pos.y + height >= mouse_y) {
			mouseDragged(mouse_x, mouse_y, flag, drag_x, drag_y);
			return true;
		}

		return false;
	}

	@Override
	public boolean mouseScrolled(double mouse_x, double mouse_y, double scroll) {
		if (!getVisible() || !getMouseInputEnabled())
			return false;

		for (var child : children) {
			var pos = child.localToScreen();

			if (child.getIgnoreMouseEventBoundaries() || pos.x <= mouse_x && pos.x + child.width >= mouse_x && pos.y <= mouse_y && pos.y + child.height >= mouse_y) {
				return child.mouseScrolled(mouse_x, mouse_y, scroll);
			}
		}

		return GuiEventListener.super.mouseScrolled(mouse_x, mouse_y, scroll);
	}

	public boolean mouseScrolledChecked(double mouse_x, double mouse_y, double scroll) {
		if (!getVisible())
			return false;

		var pos = localToScreen();

		if (getIgnoreMouseEventBoundaries() || pos.x <= mouse_x && pos.x + width >= mouse_x && pos.y <= mouse_y && pos.y + height >= mouse_y) {
			mouseScrolled(mouse_x, mouse_y, scroll);
			return true;
		}

		return false;
	}

	protected boolean keyPressedInternal(int p_94745_, int p_94746_, int p_94747_) {
		return false;
	}

	@Override
	public boolean keyPressed(int p_94745_, int p_94746_, int p_94747_) {
		if (!getVisible() || !getKeyboardInputEnabled())
			return false;

		if (!focused_internal)
			return false;

		if (focused)
			return keyPressedInternal(p_94745_, p_94746_, p_94747_);

		for (var child : children) {
			if (child.keyPressed(p_94745_, p_94746_, p_94747_)) {
				return true;
			}
		}

		return false;
	}

	protected boolean keyReleasedInternal(int p_94750_, int p_94751_, int p_94752_) {
		return false;
	}

	@Override
	public boolean keyReleased(int p_94750_, int p_94751_, int p_94752_) {
		if (!getVisible() || !getKeyboardInputEnabled())
			return false;

		if (!focused_internal)
			return false;

		if (focused)
			return keyReleasedInternal(p_94750_, p_94751_, p_94752_);

		for (var child : children) {
			if (child.keyReleased(p_94750_, p_94751_, p_94752_)) {
				return true;
			}
		}

		return false;
	}

	@Override
	public boolean charTyped(char p_94732_, int p_94733_) {
		if (!getVisible() || !getKeyboardInputEnabled())
			return false;

		for (var child : children) {
			if (child.charTyped(p_94732_, p_94733_)) {
				return true;
			}
		}

		return true;
	}

	@Override
	public boolean changeFocus(boolean p_94756_) {
		if (!getVisible())
			return false;

		return GuiEventListener.super.changeFocus(p_94756_);
	}

	@Override
	public boolean isMouseOver(double mouse_x, double mouse_y) { // called to check whenever we are hovering at this
		if (!getVisible() || !getMouseInputEnabled())
			return false;

		if (getIgnoreMouseEventBoundaries())
			return true;

		var pos = localToScreen();
		return mouse_x >= pos.x && mouse_x <= pos.x + width && mouse_y >= pos.y && mouse_y + height <= pos.y;
	}

	public void tick() {

	}
}
